import path from 'node:path'
import type { ResolvedConfig, ViteDevServer, HMRPayload, Rollup } from 'vite'
import { normalizePath } from 'vite'

export type Dict<T = unknown> = Record<string, T>

export interface OxManifest {
  namespace: string
  path: string
  requires?: string
  raw?: string
  generator?: string
}

export const PROJECT_NAME = '@open-xchange/vite-plugin-ox-manifests'

export function applyInputToOptions (input: Dict<string>, options: Rollup.InputOptions): void {
  // apply changes to input
  if (Object.keys(input).length > 0) {
    if (options.input) {
      if (typeof options.input === 'string') {
        input.index = options.input
        options.input = input
      } else if (Array.isArray(options.input)) {
        options.input = [...options.input, ...Object.values(input)]
      } else {
        Object.assign(input, options.input)
        options.input = input
      }
    }
  }
}

export function joinUrlPaths (...paths: string[]): string {
  return normalizePath(path.join(...paths))
}

export function basepath (src: string): string {
  const { dir, name } = path.parse(src)
  return joinUrlPaths(dir, name)
}

export function deepMergeObject (a: Dict, b: Dict): Dict {
  const keys = [...new Set([...Object.keys(a), ...Object.keys(b)])]
  const result = {} as Dict

  for (const key of keys) {
    if (a[key] === undefined) result[key] = b[key]
    else if (b[key] === undefined) result[key] = a[key]
    else if (Array.isArray(a[key]) !== Array.isArray(b[key])) throw new Error('Cannot merge different types')
    else if (typeof a[key] !== typeof b[key]) throw new Error('Cannot merge different types')
    else if (Array.isArray(a[key])) result[key] = [...a[key] as unknown[], ...b[key] as unknown[]]
    else if (typeof a[key] === 'object') result[key] = deepMergeObject(a[key] as Dict, b[key] as Dict)
    else result[key] = b[key]
  }

  return result
}

export function mergeManifests (manifests: OxManifest[]): OxManifest[] {
  const manifestsByHash = new Map<string, OxManifest>()
  for (const manifest of manifests) {
    const hash = `${manifest.namespace}--${manifest.path}`
    // overwrites existing ones. last wins
    manifestsByHash.set(hash, manifest)
  }
  return Array.from(manifestsByHash.values())
}

export function stringifyJSON ({ build }: ResolvedConfig, json: unknown): string {
  return (build.minify === false) ? JSON.stringify(json, null, 2) : JSON.stringify(json)
}

export function waitForReload (server: ViteDevServer): Promise<void> {
  return new Promise(resolve => {
    const send = server.ws.send.bind(server.ws)
    server.ws.send = function (arg1: string | HMRPayload, ...args: unknown[]): void {
      try {
        send(arg1 as string, ...args)
      } finally {
        if (typeof arg1 === 'object' && arg1.type === 'full-reload') resolve()
      }
    }
  })
}

export async function wait (millis: number): Promise<void> {
  await new Promise(resolve => setTimeout(resolve, millis))
}
