import { describe, it, expect } from 'vitest'
import { build, createServer } from 'vite'
import vitePluginOxExternals from '@open-xchange/vite-plugin-ox-externals'
import vitePluginOxManifests from '../../dist/index'
import { getFileFromBundle, getPort } from '../util'

const __dirname = import.meta.dirname

describe('Multiple manifest scenario', () => {
  it('works', async () => {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        write: false,
        rollupOptions: {
          input: {}
        }
      },
      plugins: [
        vitePluginOxExternals({ prefix: '$' }),
        vitePluginOxManifests({})
      ]
    })

    const manifestFile = getFileFromBundle('manifest', bundle)
    const manifests = JSON.parse(manifestFile.source)
    expect(Object.keys(manifests)).toHaveLength(2)

    expect(manifests['entry1.js'].meta.manifests[0].namespace).toBe('test')

    expect(manifests['sub/entry2.js'].meta.manifests).toHaveLength(3)
    expect(manifests['sub/entry2.js'].meta.manifests.find(m => m.namespace === 'other').namespace).toBe('other')
    expect(manifests['sub/entry2.js'].meta.manifests.find(m => m.namespace === 'test').namespace).toBe('test')
    expect(manifests['sub/entry2.js'].meta.manifests.find(m => m.namespace === 'settings').namespace).toBe('settings')
  })

  it('works in dev mode', async () => {
    const server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [
        vitePluginOxManifests({}),
        vitePluginOxExternals({ prefix: '$' })
      ]
    })
    const PORT = await getPort()
    await server.listen(PORT)

    const res = await fetch(`http://localhost:${PORT}/api/manifest.json`)
    const manifests = await res.json()
    await server.close()

    expect(manifests).toHaveLength(4)

    const entry1Index = manifests.findIndex(m => m.path.indexOf('entry1') >= 0)
    expect(manifests[entry1Index].namespace).toBe('test')
    expect(manifests[entry1Index].path).toContain('entry1')

    const otherNSIndex = manifests.findIndex(m => m.namespace === 'other')
    expect(manifests[otherNSIndex].namespace).toBe('other')
    expect(manifests[otherNSIndex].path).toContain('entry2')

    const lastIndex = manifests.findIndex(({ path, namespace }) => path.indexOf('entry2') >= 0 && namespace === 'test')
    expect(manifests[lastIndex].namespace).toBe('test')
    expect(manifests[lastIndex].path).toContain('entry2')

    const settingsIndex = manifests.findIndex(m => m.namespace === 'settings')
    expect(manifests[settingsIndex].namespace).toBe('settings')
    expect(manifests[settingsIndex].path).toContain('entry2')
  })
})
