import { createServer } from 'vite'
import { describe, it, expect, afterEach } from 'vitest'
import vitePluginProxy from '../../dist/index'
import { createTestServer, generateSelfsigned, getPort, http2Request, updateRequestManager } from '../util'

const __dirname = import.meta.dirname
process.noDeprecation = true

describe('simple proxy', () => {
  afterEach(async () => {
    updateRequestManager()
  })

  it('works', async () => {
    const server = await createTestServer()
    server.get('/api/test', (req, res) => {
      res.send({ hello: 'world' })
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        proxy: {
          '/api': {
            target: `http://localhost:${SERVER_PORT}`
          }
        }
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data, httpVersion } = await http2Request(`http://localhost:${PORT}/api/test`)
    expect(data).toEqual({ hello: 'world' })
    expect(httpVersion).toEqual('1.1')
    server1.close()
    server.close()
  })

  it('works with http2', async () => {
    const server = await createTestServer({
      http2: true,
      https: generateSelfsigned()
    })
    server.get('/api/test', (req, res) => {
      res.send({ hello: 'world' })
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      server: {
        https: generateSelfsigned()
      },
      plugins: [vitePluginProxy({
        proxy: {
          '/api': {
            target: `https://localhost:${SERVER_PORT}`,
            secure: false
          }
        }
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data, httpVersion } = await http2Request(`https://localhost:${PORT}/api/test`, {
      rejectUnauthorized: false
    })
    expect(data).toEqual({ hello: 'world' })
    expect(httpVersion).toEqual('2.0')
    server1.close()
    server.close()
  })

  it('regular expressions', async () => {
    const server = await createTestServer()
    server.get('/api/test', (req, res) => {
      res.send({ hello: 'world' })
    })
    server.get('/foo/test', (req, res) => {
      res.send({ hello: 'foo' })
    })
    server.get('/bar/test', (req, res) => {
      res.send({ hello: 'bar' })
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        proxy: {
          '^/(foo|bar)': {
            target: `http://localhost:${SERVER_PORT}`
          }
        }
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/foo/test`)
    expect(data).toEqual({ hello: 'foo' })
    const { data: data2 } = await http2Request(`http://localhost:${PORT}/bar/test`)
    expect(data2).toEqual({ hello: 'bar' })
  })

  it('rewrite rule', async () => {
    const server = await createTestServer()
    server.get('/appsuite/api/test', (req, res) => {
      res.send({ hello: 'world' })
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        proxy: {
          '/api': {
            target: `http://localhost:${SERVER_PORT}`,
            rewrite: () => '/appsuite/api/test'
          }
        }
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/api/test`)
    expect(data).toEqual({ hello: 'world' })
  })

  it('prepends a path', async () => {
    const server = await createTestServer()
    server.get('/appsuite/api/test', (req, res) => {
      res.send({ hello: 'world' })
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        proxy: {
          '/api': {
            target: `http://localhost:${SERVER_PORT}`,
            prependPath: '/appsuite'
          }
        }
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/api/test`)
    expect(data).toEqual({ hello: 'world' })
  })
})
